// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *
 * 		Link Access Protocol - a File Link
 *
 * This is a particular type of link access via UNIX file (which can be
 * a pipe, btw). The name of the file is determined from a env
 * variable LAPFILENAME
 * In addition, an env variable LAPFILEERR may be specified to tell
 * the error rate to be introduced to the input stream. E.g., if
 * LAPFILEERR=10, every 10th byte (on average) read from the file
 * would be distorted.
 *
 * $Id: LAPFile.cc,v 1.4 1999/09/24 18:10:46 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation
#endif

#include "util.h"
#include "LAPFile.h"
#include <errno.h>


static const char LAPFile_name [] = "LAPFILENAME";
static const char LAPError_rate_name [] = "LAPFILEERR";

			// Determine the name of the LAP File and open it
LAPFileOut::LAPFileOut(void)
{
  const char * file_name = xgetenv(LAPFile_name,0);
  if( (fp = fopen(file_name,"wb")) == 0 )
    perror("opening LAPFile"),
    _error("Failed to open LAPFile '%s' due to the reason above",file_name);
}

			// Done with the file
LAPFileOut::~LAPFileOut(void)
{
  fclose(fp);
}

  			// The standard LAPOut interface
void LAPFileOut::write_byte(const unsigned char a_byte)
{
  putc(a_byte,fp);
  if( ferror(fp) )
   perror("error writing LAPFile"), _error("error writing LAPFile");
}

void LAPFileOut::begin_packet(void)
{
  write_byte(SOH);
}

void LAPFileOut::finish_packet(void)
{
  fflush(fp);
}


static int get_error_rate(void)
{
  const int rate = atoi(xgetenv(LAPError_rate_name,"0"));
  assert( rate >= 0 );
  if( rate != 0 )
    Logger::log("every %d-th byte (on average) read from the file would be "
    		"mangled",rate);
  return rate == 0 ? Random::One_in_n::never : rate;
}

			// Determine the name of the LAP In File and open it
LAPFileIn::LAPFileIn(void)
	: error_event(get_error_rate())
{
  const char * file_name = xgetenv(LAPFile_name,0);
  if( (fp = fopen(file_name,"rb")) == 0 )
    perror("opening LAPFile"),
    _error("Failed to open LAPFile '%s' due to the reason above",file_name);
}

			// Done with the file
LAPFileIn::~LAPFileIn(void)
{
  fclose(fp);
}

			// (unreliable) read from the file
int LAPFileIn::get_char(void)
{
  const int c = getc(fp);
  if( c == EOF )
    return c;
  if( (bool)error_event )
    return c ^ 1;		// mangle the byte; toggle the lowest bit
  else
    return c;
}

  			// The standard LAPIn interface
unsigned char LAPFileIn::read_byte(void)
{
  const int c = get_char();
  if( c == EOF )
    throw rc_input_error(errno);
  return c;
}

void LAPFileIn::look_for_start_of_packet(void)
{
  register int c;
  while( (c=get_char()) != LAPFileOut::SOH )
    if( c == EOF )
      if( feof(fp) )
        throw rc_eof();
      else
        throw rc_input_error(errno);
}

void LAPFileIn::ending_packet(void)
{
}
