// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *
 * 			 Definition of a Segment
 *   a fragment of the file, and the logical unit of the transmission
 *
 * Every segment starts with
 *	start-of-segment char(s)
 *	(short) id of the broadcast session (uniquely related to a file)
 *	(ulong) sequence number
 *	(byte)  length of the following data (doesn't include the fields
 *		above, nor the CRC field)
 * followed by the packet data and terminated by a 16-bit CRC.
 * Note, the 'start-of-segment' character(s) are actually determined
 * by a link access layer, and are not a part of the logical segment.
 * All segments have the same size (which is given in the 0th segment)
 *
 * The 0th segment (that is, the segment with seq_number = 0) has a
 * special structure:
 *	(ulong)	file size
 *	(byte)  MTU, that is, the size of the data in all packets in this
 *		session
 *	(ulong) max seq number, that is, how many data packets to follow
 *	(chars) file name
 *
 * An env variable LAPMTU may be specified to tell the size of the
 * output packets (MTU).
 *
 * $Id: Segment.h,v 1.4 1999/09/24 20:48:24 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifndef __GNUC__
#pragma once
#else
#pragma interface
#endif

#ifndef __Segment_h
#define __Segment_h

#include "LAP.h"
#include "myenv.h"

class SegmentOutStream;
class SegmentInStream;

typedef unsigned short ID;

			// All segments start with this....
struct SegmentHeader
{
  const unsigned long seq_number;	// Sequence number
  const ID session_id;			// ID of the current session
  const unsigned char data_length;	// Length of the data only
  					// Must be within (1,255)
  					
  inline int q_header_size(void) const  // True header size, w/o any padding
  	{ return sizeof(seq_number) + sizeof(session_id) + 
  		 sizeof(data_length); }
  		 
  SegmentHeader(const unsigned long _seq_number, const short _session_id,
  	        const int _data_length)
    : seq_number(_seq_number), session_id(_session_id),
      data_length(_data_length)
    { assert( _data_length > 1 && _data_length < 255 ); }
  void write(SegmentOutStream& out_packet) const;
  void read(SegmentInStream& in_packet);
};



class Segment0 : public SegmentHeader
{
protected:
  char file_name[20];			// only the basic portion of the name, w/o path
  const unsigned long file_size;
  unsigned long max_seq_no;		// That is, how many data packets
  const unsigned char MTU;		// Max transmission unit: data_length
  					// in all the packets
  					// to follow
  inline int q_data_size(void) const
  	{ return sizeof(file_size) + sizeof(MTU) +
  	         sizeof(max_seq_no) + sizeof(file_name); }

  					// _Partially_ fill out Segment0
  					// based on what found in
  					// some_packet
  Segment0(const SegmentHeader& some_packet);
  
  static unsigned char default_MTU;
  
public:

  enum { max_MTU = 255 };		// None of the segments can be bigger
  					// than that

  Segment0(const char * full_file_name);
  void send(LAPOut& out_link) const;
  void log(const char * title) const;	// Log the contents of Segment0,
  					// which represents a session
  					// Complete a partially finished
  					// Segment0 from a packet
  void complete(SegmentInStream& packet) throw (rc_bad_packet);
  
  unsigned int q_MTU(void) const		{ return MTU; }
  unsigned int q_max_seq_no(void) const		{ return max_seq_no; }
  
  static void set_default_MTU(const int new_MTU);
  static int  get_default_MTU(void);
};


			// A stream to help compose a packet
class SegmentOutStream
{
  LAPOut& out_link;
  CRC16 crc;

  void write_byte(const unsigned char a_byte)
  { out_link.write_byte(a_byte); crc << a_byte; }

public:
  SegmentOutStream(LAPOut& _out_link);
  ~SegmentOutStream(void);		// write the CRC and flush the packet

  SegmentOutStream& operator << (const unsigned char a_byte)
  		{ write_byte(a_byte); return *this; }
  SegmentOutStream& operator << (const unsigned short a_short);
  SegmentOutStream& operator << (const unsigned long a_long);
  SegmentOutStream& operator << (const SegmentHeader& seg_header);
  
  					// Write an array of bytes
  					
  		// Array: a representation for an array. Note, for safety
  		// an Array object cannot be constructed explicitly,
  		// either on stack or on heap. An Array object can only
  		// be constructed via a friend function byte_array,
  		// and is always transient.
  class Array
  {
    friend SegmentOutStream;
    friend SegmentInStream;
    const char * const ptr;
    const unsigned long size;
    Array(const char * byte_array, const unsigned long _size) // private constructor
  	: ptr(byte_array), size(_size) {}
    Array& operator = (const Array&); // unimplemented and forbidden
    //    Array(const Array&);	      // unimplemented and forbidden
  public:
    friend inline Array byte_array(const char * ptr, const unsigned long size)
    	{ return Array(ptr,size); }
  };
  SegmentOutStream& operator << (const Array& array);
  
};

inline SegmentOutStream&
SegmentOutStream::operator << (const SegmentHeader& seg_header)
{
  seg_header.write(*this);
  return *this;
}

			// A stream to help de-compose a packet
			// When the stream is fully constructed, it contains
			// the complete packet, with CRC and other checks
			// performed
class SegmentInStream
{
  unsigned char buffer[Segment0::max_MTU+1];	// The buffer where the packet is received
  				// Two ptrs in buffer for the curent reading
  				// and writing position
  const unsigned char * read_ptr;
  unsigned char * write_ptr;
  CRC16 curr_crc;		// CRC accumulated while the packet is being received
  
				// The packet header is being read as a part
				// of the receiving process
  SegmentHeader header;
  
  				// Take the current byte from the buffer
  unsigned char take_byte(void)
  	{ assert( write_ptr > read_ptr ); return *read_ptr++; }

  				// Receive a byte from the link and into the
  				// buffer
  void receive_byte(LAPIn& in_link);
  
public:
				// Receive a packet. It throws up if there was
				// an input error
  SegmentInStream(LAPIn& in_link);
  const SegmentHeader& q_header(void) const	{ return header; }

  SegmentInStream& operator >> (unsigned char& a_byte)
  		{ a_byte = take_byte(); return *this; }
  SegmentInStream& operator >> (unsigned short& a_short);
  SegmentInStream& operator >> (unsigned long& a_long);
  SegmentInStream& operator >> (const SegmentOutStream::Array& array);
  void write(FILE * fp);	// Dump the rest if the stream into a file
};

#endif
