// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *			Main Broadcasting Module
 *
 * The purpose of this module is to transmit a file. Since the channel
 * is unrealiable, the file is shredded into the chunks that get transmitted
 * over and over again. Since the transmitter has no feedback, and no
 * knowledge how well the chunks are received, the hope is that if a
 * receiver missed a few packets, it can get them another time around.
 *
 * Synopsis:
 *	broadcaster -ssleep_interval -nrepetition_factor file-name1 ...
 * Each file in the list is broadcast 'repetition_factor' times; after
 * the whole list is finished, broadcasting continues all over again from
 * the beginning of the list, until SIGHUP is received. After each
 * broadcast, the program sleeps for 'sleep-interval' seconds. The files
 * to broadcast do not have to exist when the program starts: as the program
 * periodically scans through the list of the files, it watches for new or
 * newly modified files, and gives them priority in broadcasting. 
 * When transmitting a file, the broadcaster watches for its modification
 * timestamp and exclusive access locks. If any attempt to modify the file
 * (by some other process) was detected, the current broadcast session is
 * aborted. When the file would be picked up again for the broadcast, the
 * new session (with a new session id) would be initiated, which in turn will
 * prompt a receiver to start receiving the file anew.
 *
 * The broadcaster generally picks up files for transmission in the order
 * they are specified in the list. However, newly modified files are given
 * preference. Whenever this program finishes broadcasting a file, it
 * timestamps it (the corresponding FileToSend object, that is). The list of
 * files is scanned again, and the file with the earliest last_broadcasted
 * timestamp will be broadcast next. If there are several files with an
 * identical timestamps, the first one is picked. When scanning through the
 * list of files, this program checks to see if a file has been modified. If it
 * has, its last_broadcasted timestamp is moved back (into the past) by
 * a certain amount (new_file_timestamp_bias). Naturally that makes the
 * timestamp appear earlier, and gives this file preference as the
 * broadcaster looks for the files with the earliest timestamp.
 * This whole technique makes the list of files look like a cyclical priority
 * queue, only nothing is being moved. It favors recently modified files,
 * yet still giving all files their fair share of broadcasting time: it's
 * obvious that even if one single file gets modified very frequently, it
 * does _not_ prevent other files in the list from being broadcast (eventually).
 *
 * Note, SIGHUP signal does not kill the broadcaster right away: it merely
 * tells the broadcaster to stop after it finishes the current transmission
 * session (if it is in the process of transmitting). Sending out a file
 * through a LAPFile interface (connected to a low-speed serial port)
 * may take time, a lot of time. Thus effect of sending the HUP signal to
 * the broadcaster process may be far from immediate. Send INT or KILL
 * signal if you want to stop the broadcaster right away.
 *
 * The following env variables are used:
 *     BC_LOG_FILE
 * name of the file where all the activity is logged. If not specified,
 * cerr is assumed
 *
 * A #defined variable DEFLAP, if set, specifies the type of the Link
 * accessed used. For example, to use a TCP/IP network link, compile
 * this module with -DDEFLAP=LAPNet compilation flag. File is
 * assumed by default.
 *
 * We also try to avoid using exceptions in this module: gcc 2.7.2 on
 * HP-PA architecture does not support C++ exceptions yet.
 *
 * $Id: broadcaster.cc,v 3.2 2000/03/13 23:49:28 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation "Segment.h"
#pragma implementation "LAP.h"
#endif

#ifndef DEFLAP
#define DEFLAP LAPFile
#endif

			// Some preprocessor black magic to form an #include
			// string and names involving DEFLAP
#define _XqYq(x,y) x ## y
#define _XY(x,y) _XqYq(x,y)
#define QMAKESTR(x) #x
#define MAKESTR(x) QMAKESTR(x)
#define MAKEINC(x) x ## .h
#define DEFLAPINC _XY(DEFLAP,.h)
//#define DEFLAPINC MAKEINC(DEFLAP)
#define DEFLAPINCSTR MAKESTR(DEFLAPINC)

#include "util.h"
#include "Segment.h"
#include DEFLAPINCSTR
#include <signal.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <limits.h>
#include <unistd.h>
#include <time.h>

#ifdef __BORLANDC__
#include <io.h>
#endif

/*
 *------------------------------------------------------------------------
 *			Broadcasting a single file
 *
 * All needed information about a file is collected in a FileToSend object
 * below. The object includes a pointer to the Segment0 describing a
 * broadcast session. Generally this field caches the session left over
 * from the previous broadcast, to be used again if the file hasn't
 * changed since.
 * The pointer can be nil (that is, the file does not exist, is locked
 * or otherwise unsuitable for broadcast). Please always check for
 * the object's validity (using operator !).
 *
 */

class FileToSend
{
  const char * file_name;		// Always a heap pointer
  time_t last_modified;
  time_t last_broadcast;
  Segment0 * session;			// broadcasting session

  					// Mark the file as being unsuitable
  					// for the broadcast at the moment
  void mark_file_bad(const char * reason);


public:
  enum { time_t_plus_infinity = (time_t)(LONG_MAX),
  	 new_file_timestamp_bias = 20 };// if each file takes 1 sec to transmit,
  	 				// and the sleeping time is 1 sec,
  	 				// a newly modified file gets preference
  	 				// 10:1

  FileToSend(void);			// Default constructor, so we can have
  					// arrays of FileToSend objects
  ~FileToSend(void);
  void init(const char * file_name);	// A late constructor
  void check_file(void);		// Check if the file is suitable for
  					// the brodcast, and allocate a new
  					// session if necessary
  bool do_broadcast(LAPOut& out_link);	// Broadcast the file, return FALSE
  					// if transmission was interrupted
  					
  					// Check for a validity of the session
  bool operator ! (void) const		{ return session == 0; }
  time_t q_last_broadcast(void) const	{ return last_broadcast; }
  void dump(const char * title) const;	// Dump the contents of the object
};

			// A default constructor
			// so we can make arrays of the FileToSend objects
			// Use the init() method to finish the construction
FileToSend::FileToSend(void)
	: file_name(0),
	  last_modified(0),
	  last_broadcast(time(0)),
	  session(0)
{}

			// A late constructor, does the real initialization
void FileToSend::init(const char * _file_name)
{
  file_name = strdup(_file_name);
  check_file();
}

FileToSend::~FileToSend(void)
{
  if( file_name != 0 )
    free((char*)file_name), file_name = 0;
  if( session != 0)
    delete session;
  session = 0;
}

			// Mark the file as being unsuitable for the
			// broadcast at the moment (say, because it does
			// not exist). The current section would be
			// destroyed, if existed
void FileToSend::mark_file_bad(const char * reason)	
{
  last_modified = 0;
  last_broadcast = time_t_plus_infinity;
  if( session != 0 )
    delete session;
  session = 0;
  Logger::log("File '%s': %s. skipping...",file_name,reason);
}

			// Check if the file exists and suitable for
			// the broadcast, and allocate a new session
			// (if it hasn't existed, or the file was modified
			// after that)
void FileToSend::check_file(void)
{
  assert( file_name != 0 );
  struct stat file_stat;
  if( stat(file_name,&file_stat) != 0 )
  {
    mark_file_bad("failed to get file's stat; "
    		  "the file probably does not exist");
    return;
  }
  if( file_stat.st_size == 0 )
  {
    mark_file_bad("file has zero size");
    return;
  }
  
  			// The file didn't exist (or broadcast) before
  if( session == 0 )
  {
    last_modified = file_stat.st_mtime;
    last_broadcast = 0;
  }
  else if( file_stat.st_mtime != last_modified )
  {
    last_modified = file_stat.st_mtime;
    last_broadcast -= new_file_timestamp_bias;
    Logger::log("File '%s' appears to have been modified",file_name);
    delete session;
    session = 0;
  }
  
  if( session == 0 )
    session = new Segment0(file_name);
}

 			// Get the modification timestamp of the file
 			// It's better to use file_name rather than
 			// the fileno(fp) because the mtime of an
 			// already open file is cached...
static time_t get_mtime(const char * const file_name)
{
  struct stat file_stat;
  assert( stat(file_name,&file_stat) == 0 );
  return file_stat.st_mtime;
}

			// Checks to see if a portion of the file
			// (starting from the current offset till the
			// end of the file) is locked: the lock means
			// that some other process is being modifying file,
			// and reading the file is blocked and useless
static bool is_locked(const FILE * fp)
{
  return lockf(fileno(fp),F_TEST,0);
}

			// Broadcast the file,
			// return FALSE is the broadcast was interrupted
			// (because the file was changed in the
			// meantime)
			// The file has to be suitable for the broadcast
			// (as found by check_file) and the session has to be
			// allocated
bool FileToSend::do_broadcast(LAPOut& out_link)
{
  assert( session != 0 && file_name != 0 );

  class InputFile
  {
    FILE * fp;
    char buffer[256];			// Preallocated buffer
    const int read_block_size;
  public:
    InputFile(const char * const file_name, const unsigned int MTU)
    	: fp(fopen(file_name,"rb")),
    	  read_block_size(MTU)
    	{ if(fp == 0) perror("File open error");
    	  assert(MTU < sizeof(buffer)); }
    ~InputFile(void) { if(fp != 0) fclose(fp); }
    bool operator ! (void) const	{ return fp == 0; }
    void rewind(void)			{ assert(fp); ::rewind(fp); }
    bool is_locked(void) const	{ assert(fp); return lockf(fileno(fp),F_TEST,0); }
    int read(void)		{ assert(fp); 
    			return fread(buffer,sizeof(char),read_block_size,fp); }
    operator SegmentOutStream::Array (void)
    	{ return byte_array(buffer,read_block_size); }
  };
  
  InputFile fp(file_name,session->q_MTU());
  if( !fp )
    return mark_file_bad("failed to open the file"),
    false;
  
  session->log("Starting");
  session->send(out_link);
  session->send(out_link);  // and do it again, 0th segment is kind of important

  fp.rewind();
  for(register unsigned long i=1; i<= session->q_max_seq_no(); i++)
  {
    if( fp.is_locked() || last_modified != get_mtime(file_name) )
      return Logger::log("Session aborted: the file was changed"),
      false;

    if( fp.read() <= 0 )
      return perror("reading input file"),
      mark_file_bad("EOF or error reading an input file"),
      false;
    
    SegmentOutStream(out_link) << SegmentHeader(i,session->session_id,
    						session->q_MTU())
    			       << fp;
  }
  Logger::log("Finished broadcast in session %d",session->session_id);
  last_broadcast = time(0);
  return true;
}

				// Dump the contents of the object
void FileToSend::dump(const char * title) const
{
  Logger::log("FileToSend object %s:\n\tfile name: '%s'"
  "\n\tlast_modified %d\n\tlast_broadcast %d\n\tsession id %d",
  title, file_name == 0 ? "<not set>" : file_name, last_modified, 
  last_broadcast, session == 0 ? 0 : session->session_id);
}

/*
 *------------------------------------------------------------------------
 *			Dealing with a list of files to send
 */

class FileList
{
  FileToSend * const file_list;
  const int no_files;
public:
  FileList(const int _no_files, const char * argv[]);
  ~FileList(void);
  FileToSend& pick_file(void);
};

			// Construct the list of the files from the
			// remaining command line arguments (after all
			// command-line options have been scanned)
FileList::FileList(const int _no_files, const char * argv[])
	: file_list(new FileToSend[_no_files]),
	  no_files(_no_files)
{
  assert( no_files > 0 );
  for(register int i=0; i<no_files; i++)
    file_list[i].init(argv[i]);
}

FileList::~FileList(void)
{
  delete [] file_list;
}

			// Go through all the list and re-check every
			// file (to detect if a file has just shown up or was
			// modified)
			// At the same time, find the file with the
			// earliest broadcast time
			// Note, if all files in the list are bad,
			// the FileToSend object returned may be invalid;
			// always check (with the operator !) before using.
FileToSend& FileList::pick_file(void)
{
  int picked_file = 0;
  time_t earliest_timestamp = FileToSend::time_t_plus_infinity;
  for(register int i=0; i<no_files; i++)
  {
    FileToSend& file = file_list[i];
    file.check_file();
    if( difftime(file.q_last_broadcast(),earliest_timestamp) < 0 )
      earliest_timestamp = file.q_last_broadcast(),
      picked_file = i;
  }
  return file_list[picked_file];
}

/*
 *------------------------------------------------------------------------
 *         Parsing the input parameters and routing the computation
 */

                                // Print the help as how to use the program
static void help(const char * error_message)
{
  message("\n%s\n",error_message);
  const char * messages [] =
  { "\nBroadcaster of files over low-bandwidth communication channels",
    "\twith fragmentation\n",
    "Usage",
    "\tbroadcaster -ssleep_interval -nrepetition_factor input_file_name1 ...",
    "\nwhere\n",
    "\t-sn\ta sleep-interval, the number of seconds the program",
    "\t\tsleeps after each broadcast session",
    "\t-nn\t the number of times each file is repeatedly broadcast"
    "\nThe program broadcasts the specified files repeatedly in a loop",
    "until a SIGHUP signal is received:",
    "\tkill -HUP pid\nto stop the program gracefully.",
    "The files are broadcast in the order they are specified on the command line",
    "However, new or newly modified files are given preference.\n",
    "The program prints a log of its activity on stderr, unless an",
    "environment variable BC_LOG_FILE specifies another file",
    "\nExamples:",
    "\tbroadcaster -s10 -n2 packed.dat observations.syn"
  };
  for(register size_t i=0; i<sizeof(messages)/sizeof(messages[0]); i++)
    message("\n%s",messages[i]);
  message("\n\n");
  exit(4);
}

static int Sleep_interval = 10;
static int Repetition_factor = 2;

                                // Scan through the command parameters in
                                // search for options.
                                // Return the number of options parsed
                                // Currently only two options,
                                // -ssleep-interval and
                                // -nrepetition_factor are supported
static int parse_options(const int argc, const char * argv[])
{
  int noptions;
  for(noptions=0; noptions < argc-1; noptions++)
  {
    const char * str = argv[1+noptions];        // argv[0] is a pgm name, skip
    if( *str++ != '-' )                         // Option should begin with -
      break;

    int value;
    char * p;
    switch(*str++)
    {
      case 's':
           value = strtol(str,&p,10);
           if( str == p || value <= 0 )
             message("\nBad option value - '%s'",str-2), help("");
           Sleep_interval = value;
           break;

      case 'n':
           value = strtol(str,&p,10);
           if( str == p || value <= 0 )
             message("\nBad option value - '%s'",str-2), help("");
           Repetition_factor = value;
           break;

      default:
           message("Unknown option - '%s'",str-2);
           help("");
    }
  }
  return noptions;
}


                                // Routing module
int main(const int argc, const char * argv[])
{
  const int noptions = parse_options(argc,argv);
  if( argc < 1 + noptions + 1 )                // First 1 stands for argv[0]
    help("At least one file name (besides option flags) has got to be specified");

  Logger::log("Each file would be broadcast %d times",Repetition_factor);
  
  FileList file_list(argc-1-noptions,argv+noptions+1);
  
  class Done : Signal
  {
    bool done;
    bool handle(void)
      { Logger::log("caught signal %d, quitting",q_signal_id()); done = true;
        return true; }
  public:
    Done(const int signal_id) : Signal(signal_id), done(false) {}
    bool operator ! (void) const	{ return !done; }
  };
  
  Done done(SIGHUP);
  _XY(DEFLAP,Out) out_stream;

  for(;!done;
       Logger::log("sleeping for %d seconds before moving to a new file",
       		   Sleep_interval),sleep(Sleep_interval))
  {
    FileToSend& in_file = file_list.pick_file();
    if( !in_file )
      continue;			// nothing to send, go to sleep
    for(register int i=0; i<Repetition_factor && !done; i++)
    {
      if( !in_file.do_broadcast(out_stream) )
        break;			// A new session is necessary
      Logger::log("sleeping for %d seconds",Sleep_interval);
      sleep(Sleep_interval);
    }
  }
  Logger::log("The broadcaster is finished");
  return 0;
}
