// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Reliable Broadcast over unreliable low-bandwidth (satellite)
 *			channels with no feedback
 *			Main Receiving Module
 *
 * The purpose of this module is to reliably receive transmission(s)
 * of a file/files broadcast by application 'broadcaster'. The broadcaster
 * sends files shredded into chunks (segments) in a loop, over and over again.
 * This program is to reassemble the chunks into a file, and write it
 * out. The program can pickup transmission starting from any packet, and
 * would listen through as many loops of broadcast as necessary to fully
 * assemble the file. After the file was sucessfully received, all further
 * packets belonging to the same broadcast session are disregarded.
 * Note, a receiver cannot send absolutely _any_ feedback information to
 * the broadcaster.
 * This program can also handle interrupted broadcasts (which are purged after
 * a certain time-to-live); It can successfully handle several simultaneous
 * broadcasts (sessions) as well.
 *
 * Synopsis:
 *	receiver command
 * where 'command' is a shell command line to execute when a file is
 * sucessfully received. The command can contain several words. A character
 * sequence @f is replaced with the path of the received file, and a sequence
 * @n is replaced with file's name as was sent by the broadcaster.
 *
 * The following env variables are used:
 *     BC_LOG_FILE
 * name of the file where all the activity is logged. If not specified,
 * cerr is assumed
 *	SESSION_TTL
 * time to live for a session (in minutes). If not specified, 720 (12hrs)
 * is assumed.
 *
 * A #defined variable DEFLAP, if set, specifies the type of the Link
 * accessed used. For example, to use a TCP/IP network link, compile
 * this module with -DDEFLAP=LAPNet compilation flag. File is
 * assumed by default.
 *
 * $Id: receiver.cc,v 1.7 2001/07/27 23:36:59 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation "Segment.h"
#pragma implementation "LAP.h"
#endif

#ifndef DEFLAP
#define DEFLAP LAPFile
#endif

			// Some preprocessor black magic to form an #include
			// string and names involving DEFLAP
#define _XqYq(x,y) x ## y
#define _XY(x,y) _XqYq(x,y)
#define QMAKESTR(x) #x
#define MAKESTR(x) QMAKESTR(x)
#define MAKEINC(x) x ## .h
#define DEFLAPINC _XY(DEFLAP,.h)
//#define DEFLAPINC MAKEINC(DEFLAP)
#define DEFLAPINCSTR MAKESTR(DEFLAPINC)


#include "util.h"
#include "Segment.h"
#include DEFLAPINCSTR
#ifdef __BORLANDC__
#include <cstring.h>
#include <io.h>
static int ftruncate(const int fildes, const unsigned long length)
	{ return chsize(fildes,length); }
#else
#include <string>
#endif
#include <signal.h>
#include <time.h>

#if defined(unix)
#include <unistd.h>
#endif

#if defined(__GNUC__)
#include <bvector.h>
class bitstring : public bit_vector
{
public:
  bitstring(void) {}
  bool test(size_t i) const { return i < size() && (*this)[i]; }
  void set(size_t i) { if(i >= size()) insert(end(),2*i-size()+1,false);
  (*this)[i] = true; }
  int count(void) const {
    bit_vector::const_iterator iter = begin();
    int count = 0;
    while( iter != end() )
      if( *iter++ ) count++;
    return count;
  }
};
static inline void clear(bitstring& bs) { bs.clear(); }
static inline size_t last_bit_set(bitstring& bs)
{
  bit_vector::const_iterator iter = bs.end();
  while( iter != bs.begin() )
    if( *--iter ) return iter - bs.begin();
  return (size_t)(-1);
}
#elif defined(__BORLANDC__)
#ifdef __GNUC__
#include <BitString.h>
#else
#include "BString.h"
#endif
typedef BitString bitstring;
static inline void clear(BitString& bs) { bs.clear(); bs.right_trim(0); }
static inline size_t last_bit_set(BitString& bs) { return bs.last(1); }
#else
#include <bitstring>
static inline void clear(bitstring& bs) { bs.resize(0); }
static inline size_t last_bit_set(bitstring& bs) { return bs.rfind(1); }
#endif

static const char Session_TTL_env [] = "SESSION_TTL";

/*
 *------------------------------------------------------------------------
 *         		A single session of the broadcast
 */

class Session : protected Segment0
{
  FILE * fp;			// Place where packets are assembled
  string file_name;		// the name of that file
  const time_t time_started;
  bitstring fragments_received;
  bool to_be_killed;

  string sh_command;		// Executed when a file is received completely

  void purge(void);		// Session is finished, make a zombie
  void kill(void);		// Start from all over again
  void finish(void);		// The whole file has been received!
  void ingest_segment0(SegmentInStream& packet); // Got segment 0 at last

  				// Replace all occurences of 'pattern' in str
  				// with 'what'
  static void subst(string& str, const string& pattern, const string& what);

public:
  Session(string& _sh_command, SegmentInStream& packet);
  ~Session(void);
  time_t q_time_started(void) const	{ return time_started; }
  ID q_id(void) const			{ return session_id; }
  bool q_killed(void) const		{ return to_be_killed; }
  				// Check out if the session is basically
  				// finished
  bool q_zombie(void) const		{ return file_name.length() == 0; }
  
  				// Ingest a packet
  void ingest(SegmentInStream& packet);
};

Session::Session(string& _sh_command, SegmentInStream& packet)
	: Segment0(packet.q_header()),
	  file_name(tmpnam(0)), time_started(time(0)),
	  to_be_killed(false),
	  sh_command(_sh_command)
{
  if( (fp = fopen(file_name.c_str(),"wb")) == 0 )
    perror("Opening working file"),
    _error("Failed to open a working file '%s'",file_name.c_str());
  Logger::log("Opening a new session %d into tmp file '%s' "
  	      "upon receiving packet #%d",
  	      q_id(),file_name.c_str(),packet.q_header().seq_number);
  ingest(packet);
}

  				// Replace all occurences of pattern in str
  				// with 'what'
void Session::subst(string& str, const string& pattern, const string& what)
{
  for(size_t pos = 0; (pos = str.find(pattern,pos)) != string::npos;
      pos += what.length())
    str.replace(pos,pattern.length(),what);
}


			// The whole file has been received!
  			// Replace all occurences of @f/@n in sh_command
  			// with a file name/received name, resp.
void Session::finish(void)
{
  assert( file_size > 0 );
  fflush(fp);
  if( ftruncate(fileno(fp),file_size) != 0 )	// set the right size
    perror("ftruncate"), abort();
  fclose(fp); fp = 0;
  subst(sh_command,"@f",file_name);
  subst(sh_command,"@n",Segment0::file_name);
  Logger::log("Session %d: the whole file was just received. Executing '%s'",
  	      q_id(),sh_command.c_str());
  system(sh_command.c_str());
  purge();
}


		// The session is basically finished: the file was received
		// and handled. Make this session a zombie
void Session::purge(void)
{
  if( q_zombie() )
   return;
  if( fp != 0 )
    fclose(fp), fp = 0;
  unlink(file_name.c_str());
  file_name = "";
  clear(fragments_received);
}
		// What we've done so far wasn't of any use - start from
		// all over again...
		// We mark this session as dead, so the reaper would take
		// care of the copse
void Session::kill(void)
{
  Logger::log("session %d is now dead",q_id());  
  to_be_killed = true;
}

Session::~Session(void)
{
  purge();
}

  				// Ingest a packet
void Session::ingest(SegmentInStream& packet)
{
  if( q_zombie() )
    return;
  const SegmentHeader& header = packet.q_header();
  if( max_seq_no > 0 && header.seq_number > max_seq_no )
    throw rc_bad_packet(rc_bad_packet::inconsistent_seqno);
  
  if( fragments_received.test(header.seq_number) )
    return;			// disregard the duplicates
  
  if( header.seq_number == 0 )
    ingest_segment0(packet);
  else
  {
    if( MTU != header.data_length )
     if( fragments_received.test(0) )
       throw rc_bad_packet(rc_bad_packet::inconsistent_size);
     else
     {				// We haven't got segment0 yet, so we can't say
       kill();			// for sure what the right size is
     				// Just reset everything, and hope the next
     				// packet would clarify things...
       throw rc_bad_packet(rc_bad_packet::inconsistent_size);
     }
    fseek(fp,MTU*(header.seq_number-1),SEEK_SET);
    packet.write(fp);
  }
  fragments_received.set(header.seq_number);
  
  if( last_bit_set(fragments_received) == max_seq_no &&
      (unsigned)(fragments_received.count()) == 1 + max_seq_no )
    finish();
}

  				// Ingest Segment 0 and reconcile it
  				// with already accumulated info
void Session::ingest_segment0(SegmentInStream& packet)
{
  const unsigned char old_MTU = MTU;
  try
  {
    Segment0::complete(packet);
  }
  catch(...)
  {
    kill();
    throw;
  }
  if( (old_MTU > 0 && old_MTU != MTU) ||
      (fragments_received.count() > 0 && 
       last_bit_set(fragments_received) > max_seq_no) )
    kill();
  Segment0::log("got segment0");
}


/*
 *------------------------------------------------------------------------
 *         	Handling concurrent broadcast sessions
 */

class Sessions
{
  class SessionLink : public Session
  {
    friend class Sessions;
    SessionLink * next;
  public:
    SessionLink(string& sh_command, SegmentInStream& packet)
      : Session(sh_command,packet), next(0) {}
  };
  SessionLink * sessions;

  string sh_command;		// Executed when a file is received completely

  				// Find a session given its ID,
  				// return 0 if not found
  SessionLink * find(const ID id) const;
  
public:
  Sessions(const char * sh_command_words[]);
  ~Sessions(void);
  void remove_dead_sessions(const time_t ttl);
  void digest(SegmentInStream& packet);
};

			// Initialize the session manager
Sessions::Sessions(const char * sh_command_words[])
	: sessions(0)
{
  for(register const char ** wp = sh_command_words; *wp != 0 && **wp != '\0';
      wp++)
     sh_command += *wp, sh_command += ' ';
  Logger::log("A sh-command '%s' would be executed on each received file",
  	sh_command.c_str());
}

			// Scan the list of sessions and remove all
			// sessions whose time is up
			// ttl is the session's time to live, in
			// time_t units (secs)
void Sessions::remove_dead_sessions(const time_t ttl)
{
  const time_t time_now = time(0);
  register SessionLink *curr, *next, *prev;
  for(curr=sessions,prev=0; curr != 0; curr=next)
  {
    next = curr->next;
    if( !curr->q_killed() && curr->q_time_started() + ttl > time_now )
    {
      prev = curr;
      continue;
    }
    Logger::log("Dropping session %d, ttl %d secs is up, or the session was dead",
    		curr->q_id(),ttl);
    ( prev == 0 ? sessions : prev->next ) = next;
    delete curr;
  }
}

			// Remove all sessions
Sessions::~Sessions(void)
{
  for(register SessionLink * p = sessions; p != 0; )
  {
    SessionLink * const next = p->next;
    delete p;
    p = next;
  }
}

  				// Find a session given its ID,
  				// return 0 if not found
Sessions::SessionLink * Sessions::find(const ID id) const
{
  for(register SessionLink * sp = sessions; sp != 0; sp = sp->next)
    if( !sp->q_killed() && sp->q_id() == id )
      return sp;
  return 0;
}

			// Handle a new packet just received
void Sessions::digest(SegmentInStream& packet)
{
  SessionLink * sp;
  if( (sp = find(packet.q_header().session_id)) != 0 )
    sp->ingest(packet);
  else
  {
    sp = new SessionLink(sh_command,packet);
    sp->next = sessions;	// that won't be run if an exception was
    sessions = sp;		// thrown from the constructor... good!
  }
}

/*
 *------------------------------------------------------------------------
 *         Parsing the input parameters and routing the computation
 */

                                // Print the help as how to use the program
static void help(const char * error_message)
{
  message("\n%s\n",error_message);
  const char * messages [] =
  { "\nReceiver of file broadcasts over low-bandwidth communication "
    "channels with fragmentation\n",
    "Usage",
    "\t receiver command",
    "\nwhere\n",
    "command\t- a shell command line to execute when a file is",
    "\tsucessfully received. The command can contain several words.",
    "\tA character sequence @f is replaced with the path to the received file,",
    "\tand a sequence @n is replaced with file's name as was sent by the",
    "\tbroadcaster\n",
    "Environment variable SESSION_TTL tells the time to live for a session",
    "(in minutes). If not specified, 720 (12hrs) is assumed.\n",
    "The program prints a log of its activity on stderr, unless an",
    "environment variable BC_LOG_FILE specifies another file",
    "\nExamples:",
    "\tsetenv SESSION_TTL 30",
    "\treceiver copy @f received/@n"
  };
  for(register size_t i=0; i<sizeof(messages)/sizeof(messages[0]); i++)
    message("\n%s",messages[i]);
  message("\n\n");
  exit(4);
}

//static int Sleep_interval = 10;
                                // Scan through the command parameters in
                                // search for options.
                                // Return the no. of options parsed
                                // Currently only one option,
                                // -ssleep-interval, is supported
static int parse_options(const int argc, const char * argv[])
{
  int noptions;
  for(noptions=0; noptions < argc-1; noptions++)
  {
    const char * str = argv[1+noptions];        // argv[0] is a pgm name, skip
    if( *str++ != '-' )                         // Option should begin with -
      break;

    int value;
    char * p;
    switch(*str++)
    {
      case 's':
           value = strtol(str,&p,10);
           if( str == p )
             message("\nBad option value - '%s'",str-2), help("");
//           Sleep_interval = value;
           break;

      default:
           message("Unknown option - '%s'",str-2);
           help("");
    }
  }
  return noptions;
}

rc_bad_packet::operator const char * (void)
{
  static const char * messages[] =
   { "unknown error", "bad packet size", "CRC error", 
     "packet size inconsistent with that of others",
     "bad sequence number",
     "bad Segment 0"
   };
  if( (unsigned int)reason >= sizeof(messages)/sizeof(messages[0]) )
   return messages[0];
  else
   return messages[(int)reason];
}

                                // Routing module
int main(const int argc, const char * argv[])
{
  const int noptions = parse_options(argc,argv);
  if( argc < 1 + noptions + 1 )                // First 1 stands for argv[0]
    help("a shell command line has to be specified");

  class Done : Signal
  {
    bool done;
    bool handle(void)
      { Logger::log("caught signal %d, quitting",q_signal_id()); done = true;
        return true; }
  public:
    Done(const int signal_id) : Signal(signal_id), done(false) {}
    bool operator ! (void) const	{ return !done; }
  };

  time_t session_ttl;		// Time-to-live for a session
  if( sscanf(xgetenv(Session_TTL_env,"720"),"%ld",&session_ttl) != 1 )
    _error("Env variable %s is ill-specified. "
           "Must be a session TTL in minutes",Session_TTL_env);
  session_ttl *= 60;		// Convert to time_t units (secs)
  Logger::log("A session time-to-live is set to %d minutes",session_ttl/60);

#ifdef unix
  Done done(SIGHUP);
#else
  bool done = false;
#endif

  _XY(DEFLAP,In) in_stream;
  Sessions sessions(argv+1);		// Pass the words of the sh-command
  
  while(!done)
  {
    try{
      SegmentInStream new_packet(in_stream);
      sessions.remove_dead_sessions(session_ttl); // Ought to be in a separate thread
      sessions.digest(new_packet);
     }
     catch(rc_timeout rct)
     {
       Logger::log("Timeout");
       continue;
     }
     catch(rc_eof rct)
     {
       Logger::log("Got EOF - finishing");
       break;
     }
     catch(rc_input_error rct)
     {
       Logger::log("General input error %d",rct.io_errno);
       continue;
     }
     catch(rc_file_error rct)
     {
       Logger::log("File output error %d",rct.io_errno);
       break;
     }
     catch(rc_bad_packet rct)
     {
       Logger::log("Bad packet: %s",(const char *)rct);
       continue;
     }
     catch(...)
     {
       Logger::log("Got unidentified exception");
       throw;
     }
  }
  Logger::log("The receiver is finished");
  return 0;
}
