// This may look like C code, but it is really -*- C++ -*-
/*
 ************************************************************************
 *	Sending a file over an unreliable low-bandwidth (satellite)
 *			channel with no feedback
 *
 * The purpose of this module is to transmit a file through an unreliable
 * channel without any feedback. It's obvious there can be NO GUARANTEE
 * that the file will ever be received. Still, to maximize the chances,
 * this code shreds the file into chunks, and repeats them a few times.
 * The program uses the same transport/segmentation layers as the
 * other members of the broadcasting system family, see broadcaster.cc
 * for more details.
 *
 * Synopsis:
 *	sender -ssleep_interval -nrepetition_factor file-name
 * the given file is pushed through the communication channel
 * 'repetition_factor' times, sleeping for 'sleep_interval' seconds
 * between the successive bouts of transmission.
 *
 * As the file is being transmitted, the sender watches for its modification
 * timestamp and/or exclusive access locks. If any was detected, it is
 * assumed that the file has been changed, which will terminate
 * the program prematurely (with an exit code 1).
 * Note, the sender behaves as a glorified version of the UNIX 'cp'
 * command. That is, if the file does not exist, unreadable, etc,
 * the program simply quits.
 *
 * The following env variables are used:
 *	BC_LOG_FILE
 * name of the file where all the activity is logged. If not specified,
 * cerr is assumed
 *	LAPMTU
 * may be specified to tell the size of the transmission packet;
 * Default_MTU is assumed by default (see below in the code)
 *
 * A #defined variable DEFLAP, if set, specifies the type of the Link
 * accessed used. For example, to use a TCP/IP network link, compile
 * this module with -DDEFLAP=LAPNet compilation flag. File is
 * assumed by default.
 *
 * $Id: sender.cc,v 2.3 2000/03/13 23:48:29 oleg Exp oleg $
 *
 ************************************************************************
 */

#ifdef __GNUC__
#pragma implementation "Segment.h"
#pragma implementation "LAP.h"
#endif

#ifndef DEFLAP
#define DEFLAP LAPFile
#endif

			// Some preprocessor black magic to form an #include
			// string and names involving DEFLAP
#define _XqYq(x,y) x ## y
#define _XY(x,y) _XqYq(x,y)
#define QMAKESTR(x) #x
#define MAKESTR(x) QMAKESTR(x)
#define MAKEINC(x) x ## .h
#define DEFLAPINC _XY(DEFLAP,.h)
//#define DEFLAPINC MAKEINC(DEFLAP)
#define DEFLAPINCSTR MAKESTR(DEFLAPINC)

#ifndef Default_MTU
#define Default_MTU 240
#endif

#include "util.h"
#include "Segment.h"
#include DEFLAPINCSTR
#include <signal.h>
#include <sys/stat.h>
#if defined(unix)
#include <unistd.h>
#endif

#ifdef __BORLANDC__
#include <io.h>
#include <dos.h>
#endif

/*
 *------------------------------------------------------------------------
 *			Handling of an input file
 */

 			// Get the modification timestamp of the file
 			// It's better to use file_name rather than
 			// the fileno(fp) because the mtime of an
 			// already open file is cached...
static time_t get_mtime(const char * const file_name)
{
  struct stat file_stat;
  assert( stat(file_name,&file_stat) == 0 );
  return file_stat.st_mtime;
}

			// Checks to see if a portion of the file
			// (starting from the current offset till the
			// end of the file) is locked: the lock means
			// that some process is being modifying file,
			// and reading the file is blocked and useless
static bool is_locked(const FILE * fp)
{
#if defined(unix)
  return lockf(fileno(fp),F_TEST,0);
#else
  return false;
#endif
}

class FileBeingSent : protected Segment0
{
  FILE * fp;
  time_t last_modified;
  bool was_modified(void) const;
  const char * const full_name;
public:
  FileBeingSent(const char * const full_name);
  ~FileBeingSent(void);
  bool do_broadcast(LAPOut& out_link);
};

FileBeingSent::FileBeingSent(const char * const _full_name)
	: Segment0(_full_name),
	  fp(fopen(_full_name,"rb")),
	  full_name(strdup(_full_name))
{
  assure( fp != 0, "Failed to open an input file" );
  last_modified = get_mtime(full_name);
}

FileBeingSent::~FileBeingSent(void)
{
  free((void*)full_name);
  fclose(fp);
}

			// Check to see if the file being transmitted
			// is being (has been) modified by another
			// process
bool FileBeingSent::was_modified(void) const
{
  return is_locked(fp) || last_modified < get_mtime(full_name);
}

				// Broadcast the file,
				// return FALSE is the broadcast was interrupted
				// (because the file was changed in the
				// meantime)
bool FileBeingSent::do_broadcast(LAPOut& out_link)
{
  Segment0::log("Starting");
  Segment0::send(out_link);
  Segment0::send(out_link);  // and do it again, 0th segment is kind of important

  char buffer[256];
  assert( MTU < sizeof(buffer) );
  
  rewind(fp);
  for(register unsigned long i=1; i<= max_seq_no; i++)
  {
    if( was_modified() )
      return Logger::log("Session aborted: the file was changed"),
      false;
    if( fread(buffer,sizeof(char),MTU,fp) <= 0 )
      perror("reading input file"),
      _error("EOF or error reading an input file");
    
    SegmentOutStream(out_link) << SegmentHeader(i,session_id,MTU)
    			       << byte_array(buffer,MTU);
  }
  Logger::log("Finished broadcast in session %d",session_id);
  return true;
}

/*
 *------------------------------------------------------------------------
 *         Parsing the input parameters and routing the computation
 */

                                // Print the help as how to use the program
static void help(const char * error_message)
{
  message("\n%s\n",error_message);
  const char * messages [] =
  { "\nSending a file through a low-bandwidth communication "
    "channel with fragmentation\n",
    "Usage",
    "\tsender -ssleep_interval -nrepetition_factor input_file_name",
    "\nwhere\n",
    "\t-sn\ta sleep-interval, the number of seconds the program",
    "\t\tsleeps before starting the next loop of transmission",
    "\t-nn\tthe number of the transmission loops to perform",
    "The program prints a log of its activity on stderr, unless an",
    "environment variable BC_LOG_FILE specifies another file",
    "\nExamples:",
    "\tsender -s10 -n2 packed.dat",
    "\nNO GUARANTEE that the file will ever be received"
  };
  for(register size_t i=0; i<sizeof(messages)/sizeof(messages[0]); i++)
    message("\n%s",messages[i]);
  message("\n\n");
  exit(4);
}

static int Sleep_interval = 10;
static int Repetition_factor = 2;

                                // Scan through the command parameters in
                                // search for options.
                                // Return the number of options parsed
                                // Currently only two options,
                                // -ssleep-interval and
                                // -nrepetition_factor are supported
static int parse_options(const int argc, const char * argv[])
{
  int noptions;
  for(noptions=0; noptions < argc-1; noptions++)
  {
    const char * str = argv[1+noptions];        // argv[0] is a pgm name, skip
    if( *str++ != '-' )                         // Option should begin with -
      break;

    int value;
    char * p;
    switch(*str++)
    {
      case 's':
           value = strtol(str,&p,10);
           if( str == p || value <= 0 )
             message("\nBad option value - '%s'",str-2), help("");
           Sleep_interval = value;
           break;

      case 'n':
           value = strtol(str,&p,10);
           if( str == p || value <= 0 )
             message("\nBad option value - '%s'",str-2), help("");
           Repetition_factor = value;
           break;

      default:
           message("Unknown option - '%s'",str-2);
           help("");
    }
  }
  return noptions;
}


                                // Routing module
main(const int argc, const char * argv[])
{
  const int noptions = parse_options(argc,argv);
  if( argc != 1 + noptions + 1 )                // First 1 stands for argv[0]
    help("A file name (besides option flags) has got to be specified");

  Logger::log("The file would be broadcast %d times",Repetition_factor);

  Segment0::set_default_MTU(Default_MTU);
  
  const char * const input_file_name = argv[1+noptions];
  _XY(DEFLAP,Out) out_stream;
  
  FileBeingSent in_file(input_file_name);
  for(register int i=0; i<Repetition_factor; i++)
  {
    if( !in_file.do_broadcast(out_stream) )
      exit(1);
    Logger::log("sleeping for %d seconds",Sleep_interval);
    sleep(Sleep_interval);
  }
  Logger::log("The broadcaster is finished");
}
